package com.tzld.piaoquan.recommend.server.service.flowpool;

import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.ThreadPoolFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.stream.Collectors;

import static com.tzld.piaoquan.recommend.server.common.enums.AppTypeEnum.*;
import static com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class FlowPoolService {
    @Autowired
    private RedisTemplate<String, Integer> redisTemplate;
    @Autowired
    private FlowPoolConfigService flowPoolConfigService;

    private final String localDistributeCountFormat = "flow:pool:local:distribute:count:%s:%s";

    public final String valueFormat = "%s-%s";

    private ExecutorService pool = ThreadPoolFactory.defaultPool();

    private int[] appTypes = {VLOG.getCode(), LOVEMOVIE.getCode(), LOVELIVE.getCode(), LONGVIDEO.getCode(),
            SHORTVIDEO.getCode(), H5.getCode(), APPSPEED.getCode(), WAN_NENG_VIDEO.getCode(),
            LAO_HAO_KAN_VIDEO.getCode(), ZUI_JING_QI.getCode(), PIAO_QUAN_VIDEO_PLUS.getCode(), JOURNEY.getCode()};


    public Map<Long, Integer> getLocalDistributeCountWithLevel(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return Collections.emptyMap();
        }

        Map<Long, Integer> result = getDistributeCount(videoFlowPoolMap);


        // 处理脏数据：分发数<0
        Map<Long, String> dirties = videoFlowPoolMap.entrySet().stream()
                .filter(e -> result.get(e.getKey()) <= 0)
                .collect(Collectors.toMap(
                        e -> e.getKey(),
                        e -> e.getValue()
                ));
        asyncDelLocalDistributeCountWithLevel(dirties);

        return result;
    }

    private Map<Long, Integer> getDistributeCount(Map<Long, String> videoFlowPoolMap) {
        // 为了保证有序
        List<Map.Entry<Long, String>> entries =
                videoFlowPoolMap.entrySet().stream().sorted().collect(Collectors.toList());

        List<String> keys = entries.stream()
                .map(v -> String.format(localDistributeCountFormat, v.getKey(), v.getValue()))
                .collect(Collectors.toList());
        List<Integer> counts = redisTemplate.opsForValue().multiGet(keys);

        Map<Long, Integer> result = new HashMap<>();
        for (int i = 0; i < entries.size(); i++) {
            result.put(entries.get(i).getKey(), counts.get(i));
        }
        return result;
    }

    private void asyncDelLocalDistributeCountWithLevel(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return;
        }
        pool.execute(() -> {
            List<String> keys = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(localDistributeCountFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            redisTemplate.delete(keys);

            Map<String, Double> levelWeight = flowPoolConfigService.getLevelWeight();
            List<String> values = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(valueFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            for (String level : levelWeight.keySet()) {
                for (int appType : appTypes) {
                    redisTemplate.opsForSet().remove(String.format(KEY_WITH_LEVEL_FORMAT, appType, level), values);
                    redisTemplate.opsForSet().remove(String.format(KEY_QUICK_WITH_LEVEL_FORMAT, appType), values);
                }
            }
        });
    }

    public Map<Long, Integer> getLocalDistributeCountWithLevelScore(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return Collections.emptyMap();
        }

        Map<Long, Integer> result = getDistributeCount(videoFlowPoolMap);
        // 处理脏数据：分发数<0
        Map<Long, String> dirties = videoFlowPoolMap.entrySet().stream()
                .filter(e -> result.get(e.getKey()) <= 0)
                .collect(Collectors.toMap(
                        e -> e.getKey(),
                        e -> e.getValue()
                ));
        asyncDelLocalDistributeCountWithLevelScore(dirties);

        return result;
    }

    private void asyncDelLocalDistributeCountWithLevelScore(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return;
        }
        pool.execute(() -> {
            List<String> keys = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(localDistributeCountFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            redisTemplate.delete(keys);

            Map<String, Double> levelWeight = flowPoolConfigService.getLevelWeight();
            List<String> values = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(valueFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            for (String level : levelWeight.keySet()) {
                for (int appType : appTypes) {
                    redisTemplate.opsForZSet().remove(String.format(KEY_WITH_LEVEL_SCORE_FORMAT, appType, level), values);
                    redisTemplate.opsForZSet().remove(String.format(KEY_QUICK_WITH_LEVEL_SCORE_FORMAT, appType), values);
                }
            }
        });
    }

    public Map<Long, Integer> getLocalDistributeCountWithScore(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return Collections.emptyMap();
        }
        Map<Long, Integer> result = getDistributeCount(videoFlowPoolMap);
        // 处理脏数据：分发数<0
        Map<Long, String> dirties = videoFlowPoolMap.entrySet().stream()
                .filter(e -> result.get(e.getKey()) <= 0)
                .collect(Collectors.toMap(
                        e -> e.getKey(),
                        e -> e.getValue()
                ));
        asyncDelLocalDistributeCountWithScore(dirties);

        return result;
    }

    private void asyncDelLocalDistributeCountWithScore(Map<Long, String> videoFlowPoolMap) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return;
        }
        pool.execute(() -> {
            List<String> keys = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(localDistributeCountFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            redisTemplate.delete(keys);

            Map<String, Double> levelWeight = flowPoolConfigService.getLevelWeight();
            List<String> values = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(valueFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            for (String level : levelWeight.keySet()) {
                for (int appType : appTypes) {
                    redisTemplate.opsForZSet().remove(String.format(KEY_WITH_SCORE_FORMAT, appType, level), values);
                    redisTemplate.opsForZSet().remove(String.format(KEY_QUICK_WITH_SCORE_FORMAT, appType), values);
                }
            }
        });
    }

    public void updateLocalDistributeCountWithLevel(List<Video> videos) {
        if (CollectionUtils.isEmpty(videos)) {
            return;
        }
        Map<Long, String> removeMap = updateDistributeCount(videos);

        asyncDelLocalDistributeCountWithLevel(removeMap);

    }

    private Map<Long, String> updateDistributeCount(List<Video> videos) {
        Map<Long, String> removeMap = new HashMap<>();
        videos.stream().forEach(v -> {
            String key = String.format(localDistributeCountFormat, v.getVideoId(), v.getFlowPool());
            Long count = redisTemplate.opsForValue().decrement(key);
            if (count <= 0) {
                removeMap.put(v.getVideoId(), v.getFlowPool());
            }
        });
        return removeMap;
    }

    public void updateLocalDistributeCountWithLevelScore(List<Video> videos) {
        if (CollectionUtils.isEmpty(videos)) {
            return;
        }
        Map<Long, String> removeMap = updateDistributeCount(videos);

        asyncDelLocalDistributeCountWithLevelScore(removeMap);
    }

    public void updateLocalDistributeCount(List<Video> videos) {
        if (CollectionUtils.isEmpty(videos)) {
            return;
        }
        Map<Long, String> removeMap = updateDistributeCount(videos);

        asyncDelLocalDistributeCountWithScore(removeMap);
    }
}
